// vid_sortix.c -- sortix video driver

#include <sys/keycodes.h>
#include <sys/termmode.h>

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>

#include <dispd.h>

#include "quakedef.h"
#include "d_local.h"

viddef_t vid; // global video state
unsigned short d_8to16table[256];

int VGA_width, VGA_height, VGA_rowbytes, VGA_bufferrowbytes = 0;
byte* VGA_pagebase;

static qboolean mouse_avail;
static float mouse_x, mouse_y;
static int mouse_oldbuttonstate = 0;

// No support for option menus
void (*vid_menudrawfn)(void) = NULL;
void (*vid_menukeyfn)(int key) = NULL;

uint32_t current_palette[256];
struct dispd_session* session;
struct dispd_window* window;
struct dispd_framebuffer* fb;

uint8_t* image;
int image_xres;
int image_yres;
int image_pitch;

static uint32_t MakeColor(uint8_t r, uint8_t g, uint8_t b)
{
	return b << 0 | g << 8 | r << 16;
}

void VID_SetPalette(unsigned char* palette)
{
	size_t i;
	for ( i = 0; i < 256; i++ )
	{
		uint8_t r, g, b;
		r = *palette++;
		g = *palette++;
		b = *palette++;
		current_palette[i] = MakeColor(r, g, b);
	}
}

void VID_ShiftPalette(unsigned char *palette)
{
	VID_SetPalette(palette);
}

void VID_Init(unsigned char* palette)
{
	int pnum, chunk;
	byte *cache;
	int cachesize;
	uint8_t video_bpp;
	uint16_t video_w, video_h;
	uint32_t flags;

	int fake_argc = 0;
	char** fake_argv = { NULL };

	if ( !dispd_initialize(&fake_argc, &fake_argv) )
		Sys_Error("couldn't initialize dispd library");
	if ( !(session = dispd_attach_default_session()) )
		Sys_Error("couldn't attach to dispd default session");
	if ( !dispd_session_setup_game_rgba(session) )
		Sys_Error("couldn't setup dispd rgba session");
	if ( !(window = dispd_create_window_game_rgba(session)) )
		Sys_Error("couldn't create dispd rgba window");
	if ( !(fb = dispd_begin_render(window)) )
		Sys_Error("unable to begin rendering dispd window");

	image_xres = dispd_get_framebuffer_width(fb);
	image_yres = dispd_get_framebuffer_height(fb);
	image_pitch = image_xres * sizeof(uint8_t);

	if ( !(image = (uint8_t*) malloc(image_pitch * image_xres)) )
		Sys_Error("Couldn't allocate image buffer");

	// Set up display mode (width and height)
	vid.width = image_xres;
	vid.height = image_yres;
	vid.maxwarpwidth = WARP_WIDTH;
	vid.maxwarpheight = WARP_HEIGHT;

#if 0
	if ((pnum=COM_CheckParm("-winsize")))
	{
		if (pnum >= com_argc-2)
			Sys_Error("VID: -winsize <width> <height>\n");
		vid.width = Q_atoi(com_argv[pnum+1]);
		vid.height = Q_atoi(com_argv[pnum+2]);
		if (!vid.width || !vid.height)
			Sys_Error("VID: Bad window width/height\n");
	}
#endif

	VID_SetPalette(palette);

	// now know everything we need to know about the buffer
	VGA_width = vid.conwidth = vid.width;
	VGA_height = vid.conheight = vid.height;
	vid.aspect = ((float)vid.height / (float)vid.width) * (320.0 / 240.0);
	vid.numpages = 1;
	vid.colormap = host_colormap;
	vid.fullbright = 256 - LittleLong (*((int *)vid.colormap + 2048));
	VGA_pagebase = vid.buffer = image;
	VGA_rowbytes = vid.rowbytes = image_pitch;
	vid.conbuffer = vid.buffer;
	vid.conrowbytes = vid.rowbytes;
	vid.direct = 0;

	// allocate z buffer and surface cache
	chunk = vid.width * vid.height * sizeof(*d_pzbuffer);
	cachesize = D_SurfaceCacheForRes(vid.width, vid.height);
	chunk += cachesize;
	d_pzbuffer = Hunk_HighAllocName(chunk, "video");
	if ( !d_pzbuffer )
		Sys_Error("Not enough memory for video mode\n");

	// initialize the cache memory
	cache = (byte*) d_pzbuffer + vid.width * vid.height * sizeof(*d_pzbuffer);
	D_InitCaches (cache, cachesize);
}

void VID_Shutdown(void)
{
	dispd_destroy_window(window);
	dispd_detach_session(session);
}

static UpdateRect(vrect_t* rect)
{
	int y, x;
	uint8_t* out_frame = dispd_get_framebuffer_data(fb);
	size_t out_pitch = dispd_get_framebuffer_pitch(fb);
	for ( y = rect->y; y < rect->y + rect->height; y++ )
	{
		uint8_t* in_line = image + image_pitch * y;
		uint32_t* out_line = (uint32_t*) (out_frame + out_pitch * y);
		for ( x = rect->x; x < rect->x + rect->width; x++ )
			out_line[x] = current_palette[in_line[x]];
	}
}

void VID_Update (vrect_t *rects)
{
	while ( rects )
	{
		UpdateRect(rects);
		rects = rects->pnext;
	}
	dispd_finish_render(fb);
	if ( !(fb = dispd_begin_render(window)) )
		Sys_Error("unable to begin rendering dispd window");
}

/*
================
D_BeginDirectRect
================
*/
void D_BeginDirectRect(int x, int y, byte* pbitmap, int width, int height)
{
	// TODO: This should be erased by D_EndDirectRect!
    uint8_t* offset;
    if ( !image )
		return;
    if ( x < 0 )
		x = image_xres+x-1;
    offset = image + y * image_pitch + x;
    while ( height-- )
    {
        memcpy(offset, pbitmap, width);
        offset += image_pitch;
        pbitmap += width;
    }
}


/*
================
D_EndDirectRect
================
*/
void D_EndDirectRect(int x, int y, int width, int height)
{
    if (!image)
		return;
    if (x < 0)
		x = image_xres+x-1;
#if 0
    SDL_UpdateRect(screen, x, y, width, height);
#endif
}


/*
================
Sys_SendKeyEvents
================
*/

static int TranslateKey(int abskbkey)
{
	switch ( abskbkey )
	{
	case KBKEY_ESC: return K_ESCAPE;
	case KBKEY_ENTER: return K_ENTER;
	case KBKEY_BKSPC: return K_BACKSPACE;
	case KBKEY_TAB: return K_TAB;
	case KBKEY_SPACE: return K_SPACE;
	case KBKEY_LEFT: return K_LEFTARROW;
	case KBKEY_RIGHT: return K_RIGHTARROW;
	case KBKEY_UP: return K_UPARROW;
	case KBKEY_DOWN: return K_DOWNARROW;
	case KBKEY_LCTRL: case KBKEY_RCTRL: return K_CTRL;
	case KBKEY_LSHIFT: case KBKEY_RSHIFT: return K_SHIFT;
	case KBKEY_F1: return K_F1;
	case KBKEY_F2: return K_F2;
	case KBKEY_F3: return K_F3;
	case KBKEY_F4: return K_F4;
	case KBKEY_F5: return K_F5;
	case KBKEY_F6: return K_F6;
	case KBKEY_F7: return K_F7;
	case KBKEY_F8: return K_F8;
	case KBKEY_F9: return K_F9;
	case KBKEY_F10: return K_F10;
	case KBKEY_INSERT: return K_INS;
	case KBKEY_DELETE: return K_DEL;
	case KBKEY_PGUP: return K_PGUP;
	case KBKEY_PGDOWN: return K_PGDN;
	case KBKEY_HOME: return K_HOME;
	case KBKEY_END: return K_END;
	// TODO: K_PAUSE
	case KBKEY_NUM1: return '1';
	case KBKEY_NUM2: return '2';
	case KBKEY_NUM3: return '3';
	case KBKEY_NUM4: return '4';
	case KBKEY_NUM5: return '5';
	case KBKEY_NUM6: return '6';
	case KBKEY_NUM7: return '7';
	case KBKEY_NUM8: return '8';
	case KBKEY_NUM9: return '9';
	case KBKEY_NUM0: return '0';
	case KBKEY_A: return 'a';
	case KBKEY_B: return 'b';
	case KBKEY_C: return 'c';
	case KBKEY_D: return 'd';
	case KBKEY_E: return 'e';
	case KBKEY_F: return 'f';
	case KBKEY_G: return 'g';
	case KBKEY_H: return 'h';
	case KBKEY_I: return 'i';
	case KBKEY_J: return 'j';
	case KBKEY_K: return 'k';
	case KBKEY_L: return 'l';
	case KBKEY_M: return 'm';
	case KBKEY_N: return 'n';
	case KBKEY_O: return 'o';
	case KBKEY_P: return 'p';
	case KBKEY_Q: return 'q';
	case KBKEY_R: return 'r';
	case KBKEY_S: return 's';
	case KBKEY_T: return 't';
	case KBKEY_U: return 'u';
	case KBKEY_V: return 'v';
	case KBKEY_W: return 'w';
	case KBKEY_X: return 'x';
	case KBKEY_Y: return 'y';
	case KBKEY_Z: return 'z';
	// TODO: More stuff here!
	}
	return 0;
}

void Sys_SendKeyEvents(void)
{
	const unsigned termmode = TERMMODE_KBKEY
	                        | TERMMODE_UNICODE
	                        | TERMMODE_SIGNAL
	                        | TERMMODE_NONBLOCK;
	if ( settermmode(0, termmode) )
		Sys_Error("Problem setting terminal mode");

	uint32_t codepoint;
	ssize_t numbytes;
    while ( 0 < (numbytes = read(0, &codepoint, sizeof(codepoint))) )
    {
		int kbkey = KBKEY_DECODE(codepoint);
		int abskbkey = (kbkey < 0) ? -kbkey : kbkey;
		int key = TranslateKey(abskbkey);
		Key_Event(key, 0 <= kbkey);
    }
}

void IN_Init (void)
{
    if ( COM_CheckParm ("-nomouse") )
        return;
    mouse_x = mouse_y = 0.0;
    mouse_avail = 1;
}

void IN_Shutdown (void)
{
    mouse_avail = 0;
}

void IN_Commands (void)
{
    int i;
    int mouse_buttonstate;

    if (!mouse_avail) return;
#if 0
    i = SDL_GetMouseState(NULL, NULL);
    /* Quake swaps the second and third buttons */
    mouse_buttonstate = (i & ~0x06) | ((i & 0x02)<<1) | ((i & 0x04)>>1);
    for (i=0 ; i<3 ; i++) {
        if ( (mouse_buttonstate & (1<<i)) && !(mouse_oldbuttonstate & (1<<i)) )
            Key_Event (K_MOUSE1 + i, true);

        if ( !(mouse_buttonstate & (1<<i)) && (mouse_oldbuttonstate & (1<<i)) )
            Key_Event (K_MOUSE1 + i, false);
    }
    mouse_oldbuttonstate = mouse_buttonstate;
#endif
}

void IN_Move (usercmd_t *cmd)
{
    if (!mouse_avail)
        return;

    mouse_x *= sensitivity.value;
    mouse_y *= sensitivity.value;

    if ( (in_strafe.state & 1) || (lookstrafe.value && (in_mlook.state & 1) ))
        cmd->sidemove += m_side.value * mouse_x;
    else
        cl.viewangles[YAW] -= m_yaw.value * mouse_x;
    if (in_mlook.state & 1)
        V_StopPitchDrift ();

    if ( (in_mlook.state & 1) && !(in_strafe.state & 1)) {
        cl.viewangles[PITCH] += m_pitch.value * mouse_y;
        if (cl.viewangles[PITCH] > 80)
            cl.viewangles[PITCH] = 80;
        if (cl.viewangles[PITCH] < -70)
            cl.viewangles[PITCH] = -70;
    } else {
        if ((in_strafe.state & 1) && noclip_anglehack)
            cmd->upmove -= m_forward.value * mouse_y;
        else
            cmd->forwardmove -= m_forward.value * mouse_y;
    }
    mouse_x = mouse_y = 0.0;
}

/*
================
Sys_ConsoleInput
================
*/
char *Sys_ConsoleInput (void)
{
    return 0;
}
